#include <registers.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

const char _IMAGE_DESC[] = "Linux Image Loader";

void _excUartInit();
void _excUartOut(const char*);

const bool __RELEASE__ = FALSE;

void hardwareInit() {
}

__attribute__((format(printf,1,2)))
static int outf(const char * format, ...) {
	char buf[1024];
	va_list ap;
	int ret;
	
	va_start(ap, format);
	ret = vsnprintf(buf, sizeof(buf), format, ap);
	
	/* some hack: use the uart output of the exception handler */
	_excUartOut(buf);
	
	return ret;
}

extern char kerneldata;
extern char kerneldata_end;
extern char initrddata;
extern char initrddata_end;

void linuxSetupParameters(void * loc, uint32_t ramStart, uint32_t ramLen,
	uint32_t initrdLoad, uint32_t initrdLen);

__attribute__((section(".ramtext"),noreturn,long_call,noinline))
static void startKernel();

__attribute__((noreturn))
int main() {
	/* some hack: use the uart initialization of the exception handler */
	_excUartInit();
	
	uint32_t BASE = 0xa0000000;
	uint32_t KERNEL = BASE + 0x8000;
	uint32_t PARAMS = BASE + 0x100;
	uint32_t INITRD = BASE + 0x800000;
	
	uint32_t kernellen = (uint32_t)&kerneldata_end - (uint32_t)&kerneldata;
	uint32_t initrdlen = (uint32_t)&initrddata_end - (uint32_t)&initrddata;
	
	/* load kernel */
	outf("Loading Kernel from %08x to %08x. Size: %d bytes\r\n",
		(uint32_t)&kerneldata, KERNEL, kernellen);
	memcpy((void*)KERNEL, (void*)&kerneldata, kernellen);
	/* load initrd */
	outf("Loading Initrd from %08x to %08x. Size: %d bytes\r\n",
		(uint32_t)&initrddata, INITRD, initrdlen);
	memcpy((void*)INITRD, (void*)&initrddata, initrdlen);
	/* setup kernel */
	outf("Setting parameters at %08x\r\n", (uint32_t)PARAMS);
	linuxSetupParameters((void*)PARAMS, BASE, 32 * 1024 * 1024,
		INITRD, initrdlen);
	/* disable core hardware and jump */
	outf("Disabling Core Hardware and jumping into kernel...\r\n");
	startKernel();
}

__attribute__((interrupt ("IRQ")))
void hplarmv_irq() {}

__attribute__((section(".ramtext"),noreturn,long_call,noinline))
static void startKernel() {
	asm volatile(
		"ldr ip, =(1 << 12 | 1 << 11 | 1 << 2 | 1 << 0)\n\t"
		/* unlock icache */
		"mcr p15, 0, r0, c9, c1, 1\n\t"
		/* unlock dcache */
		"mcr p15, 0, r0, c9, c2, 1\n\t"
		/* unlock itlb */
		"mcr p15, 0, r0, c10, c4, 1\n\t"
		/* unlock dtlb */
		"mcr p15, 0, r0, c10, c8, 1\n\t"
		/* deactivate mmu, btb and caches */
		"mrc p15, 0, r0, c1, c0, 0\n\t"
		"bic r0, ip\n\t"
		"mcr p15, 0, r0, c1, c0, 0\n\t"
		/* CPWAIT */
		"mrc p15, 0, r0, c2, c0, 0\n\t"
		"mov r0, r0\n\t"
		"sub pc, #0x4\n\t"
		/* invalidate icache */
		"mcr p15, 0, r0, c7, c5, 0\n\t"
		/* invalidate dcache */
		"mcr p15, 0, r0, c7, c6, 0\n\t"
		/* invalidate btb */
		"mcr p15, 0, r0, c7, c5, 6\n\t"
		/* invalidate tlb */
		"mcr p15, 0, r0, c8, c7, 0\n\t"
		/* setup registers for kernel */
		"mov r0, #0\n\t"
		"ldr r1, =775\n\t"
		"mov ip, #0xa0000000\n\t"
		"add r2, ip, #0x100\n\t"
		/* jump */
		"add pc, ip, #0x8000\n\t");
	while(1);
}

